#include <iostream>
#include <fstream>
#include <stdexcept>
#include <sstream>
#include <vector>

#include <boost/timer.hpp>
#include <boost/progress.hpp>
#include <boost/scoped_ptr.hpp>
#include <boost/lexical_cast.hpp>
#include <boost/filesystem/path.hpp>
#include <boost/filesystem/fstream.hpp>

#include <outilex/text_fsa.h>
#include <outilex/sentence_fsa.h>
#include <outilex/transduction-new.h>
#include <outilex/deter_string_fst.h>

#include <outilex/usage.h>


using namespace std;
using namespace boost;
namespace fs = boost::filesystem;



const char * USAGE_DESCRIPTION =
"usage: fst-transduct -l <lingdef> -fst <fst> [-n <nbiter>][-html|-txt][-concord|-trans][-m|-r|-i][-o <outputres>] <txtfsa>\n";


enum { TRANSDUCTION, CONCORDANCE };


int apply_transducer(itext_fsa & itext, deter_string_fst & fst, ostream & os,
                     int MODE, int transmode, int outputmode) {

  //cerr << "HERE\n";
  sentence_fsa fsa;

  int sentenceno = 0, totalmatches = 0;
  boost::timer tmr;
  int total = itext.size();
  progress_display show_progress(total, cout);

  while (itext >> fsa) {

    //cerr << "sentence #" << sentenceno <<  "/" << total << endl;

    if (! fsa.empty()) {

      if (MODE == TRANSDUCTION) {

        totalmatches += transduction2(fsa, fst, os, transmode, outputmode);

      } else { // concordance mode : only output sentences with matching sequences

        ostringstream oss;
        int nbmatches = transduction2(fsa, fst, oss, transmode, outputmode);
        if (nbmatches) { os << oss.str(); }

        totalmatches += nbmatches;
      }
    }

    ++sentenceno;
    ++show_progress;
  }

  cout << "done. " << sentenceno << " sentences parsed.\n"
    << "found " << totalmatches << " matching sequences, " << tmr.elapsed() << "s.\n";

  if (outputmode == HTML_OUTPUT) { os << "<br>\n" << totalmatches << " matching sequences\n"; }

  return totalmatches;
}



int main(int argc, char ** argv) try {

  fs::path txtpath, lingdefpath, fstpath, outputpath;

  int transmode  = MERGE_MODE;
  int outputmode = TEXT_OUTPUT;
  int nbiter = 1;

  int MODE = TRANSDUCTION;

  {
    char * text = getenv("LINGDEF");
    if (text) {
      lingdefpath = fs::path(text);
    }
  }

  argv++, argc--;
  if (argc == 0) { usage(); }

  while (argc) {
    
    string arg = *argv;
    
    if (arg == "-l") {
      
      argv++, argc--;
      if (argc == 0) { bad_args(); }
      lingdefpath = fs::path(*argv);
    
    } else if (arg == "-o") {
      
      argv++, argc--;
      if (argc == 0) { bad_args(); }
      outputpath = fs::path(*argv);
    
    } else if (arg == "-fst") {
      
      argv++, argc--;
      if (argc == 0) { bad_args(); }
      fstpath = fs::path(*argv);
    
    } else if (arg == "-html") {
    
      outputmode = HTML_OUTPUT;
    
    } else if (arg == "-text" || arg == "-txt") {
    
      outputmode = TEXT_OUTPUT;
    
    } else if (arg == "-concord") {
    
      MODE = CONCORDANCE;
    
    } else if (arg == "-trans") {
    
      MODE = TRANSDUCTION;
    
    } else if (arg == "-m") {
    
      transmode = MERGE_MODE;
    
    } else if (arg == "-r") {
    
      transmode = REPLACE_MODE;
    
    } else if (arg == "-i") {
    
      transmode = IGNORE_MODE;
    
    } else if (arg == "-h") {
    
      usage();
    
    } else if (arg == "-n") {

      argv++, argc--;
      if (argc == 0) { bad_args(); }

      nbiter = boost::lexical_cast<int>(*argv);

    } else {

      txtpath = fs::path(arg);
    }

    argv++, argc--;
  }

  if (txtpath.empty() || lingdefpath.empty() || fstpath.empty()) { bad_args(); }

  if (outputpath.empty()) {

    string resname;
    if (MODE == TRANSDUCTION) {
      resname = "transduction-res";
    } else {
      resname = "concord";
    }

    if (outputmode == HTML_OUTPUT) {
      resname += ".html";
    } else {
      resname += ".txt";
    }
    outputpath = resname;
  }
  
  fs::ofstream os(outputpath);
  if (! os) { cerr << "unable to open " << outputpath.string() << endl; exit(1); }

  cout << nbiter << " iterations\n";

  if (outputmode == HTML_OUTPUT) {
    os << 
      "<html>\n"
      "<head>\n"
      "<title>"
      << txtpath.leaf() << " : " <<
      ((MODE == TRANSDUCTION) ? "Transduction by " : "Matches with ") << fstpath.leaf() <<
      "</title>\n" <<
      "<meta http-equiv='Content-Type' content='text/html; charset=utf-8'>\n"
      "</head>\n"
      "<body bgcolor='#ffffff' text='#000000'>\n"
      "<h1>" << txtpath.leaf() << "</h1>\n";
  }


  ling_def lingdef(lingdefpath);

  scoped_ptr<itext_fsa> p_itext(new_itext_fsa(txtpath, & lingdef));
  itext_fsa & itext = *p_itext;

  //  deter_string_fst fst(fstpath, & lingdef);

  string_transducer T(fstpath, & lingdef);
  deter_string_fst fst(T);

  for (int i = 0; i < nbiter; ++i) {
    cout << "iteration #" << i << "\n";
    itext.rewind();
    apply_transducer(itext, fst, os, MODE, transmode, outputmode);
  }


  {
    fs::path ofstpath = fstpath.branch_path() / (fstpath.leaf() + ".new");
    cout << "NOT saving partially determinized transducer in " << ofstpath.string() << endl;
    //fst.write(ofstpath);
  }


  if (outputmode == HTML_OUTPUT) { os << "</body>\n</html>\n"; }

  cout << "result in " << outputpath.string() << ".\n";
  return 0;

} catch (exception & e) {

    cerr << "exception caught :" << e.what() << endl; exit(1);
  
} catch (...) { cerr << "caught an OVNI?\n"; exit(1); }

